#!/usr/bin/python3
#
#@file polyservice.py
#@brief python module to run a command line utility for a protocol
#@author Jason Berger
#@date 02/19/2019
#

import sys
import pkgutil
import subprocess
from polypacket.protocol import *
from collections import deque
from cobs import cobs
import struct
import random
import threading
import errno
import socket
import serial






def packVarSize(value):
    bytes = bytearray([])
    tmp =0
    size =0

    while( value > 0):
        tmp = value & 0x7F
        value = value >> 7
        if value > 0:
            tmp |= 0x80
        bytes.append(tmp)

    if len(bytes) == 0:
        bytes.append(0)

    return bytes

def readVarSize(bytes):
    tmp =0
    val =0
    size =0

    for i in range(4):
        tmp = bytes[i]
        size+= 1
        val = val << 7
        val |= tmp & 0x7F
        if (tmp & 0x80) == 0:
            break

    return int(val), int(size)


class PolySerial (threading.Thread):
    def __init__(self, iface, port, baud ):
        threading.Thread.__init__(self)
        self.iface = iface
        self.port = port
        self.baud = baud
        self.opened = False
        try:
            self.serialPort = serial.Serial(
                port = port,
                baudrate=baud,
                parity=serial.PARITY_NONE,
                stopbits=serial.STOPBITS_ONE,
                bytesize= serial.EIGHTBITS
            )
            self.opened = True
            self.iface.print(self.iface.name + " Port Opened : " + port)
        except serial.SerialException as e:
            print(e)
            self.iface.print(self.iface.name + " Could not open port : " + port)


    def __del__(self):
        self.close()
        self.join()     #stop thread


    def close(self):
        if self.opened:
            self.serialPort.close()


    def send(self, data):
        if self.opened:
            self.serialPort.write(data)

    def run(self):
        if self.opened:
            while True:
                if self.serialPort.inWaiting() > 0:
                    data = self.serialPort.read()
                    self.iface.feedEncodedBytes(data)

class PolyTcp (threading.Thread):
    def __init__(self, iface, localPort ):
        threading.Thread.__init__(self)
        self.iface = iface
        self.localPort = localPort
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.host = 0
        self.mode = 'server'
        self.connection = None 
        self.client_address = None

    def __del__(self):
        self.socket.close()
        self.join()

    def close(self):
        self.socket.close()

    def connect(self, hostIp, hostPort):
        self.iface.print(" TCP Connecting to " + hostIp + ":"+str(hostPort))
        self.host = (hostIp, hostPort)
        self.mode = 'client'
        try:
            self.socket.connect(self.host)
        except Exception  as e:
            self.iface.print(str(e))
    
    def listen(self):
        try:
            self.socket.bind(('', self.localPort))
            self.socket.listen(1)
            self.iface.print(" TCP Listening on port: " + str(self.socket.getsockname()[1]))
        except Exception  as e:
            self.iface.print(str(e))

    def send(self, data):
        try:
            # self.iface.print(str(self.host))
            # self.iface.print(" >>> " + ''.join(' {:02x}'.format(x) for x in data))
            if self.mode == 'server':
                self.connection.sendall(data)
            else:
                self.socket.sendall(data)
        except Exception  as e:
            self.iface.print(str(e))

    def run(self):

        if self.mode == 'server':
            while True:
                self.connection, self.client_address = self.socket.accept()
                self.iface.service.print(" Connection Accepted: " + str(self.client_address))
                while True:
                    try:
                        data = self.connection.recv(1024)
                        if data:
                            self.iface.feedEncodedBytes(data)
                        else:
                            break
                    except IOError as e:  # and here it is handeled
                        self.iface.service.print(str(e))
                        break
                self.iface.service.print(" TCP Disconnected")
                self.connection.close()
        else :#client
            while True:
                try:
                    data = self.socket.recv(1024)
                    if data:
                        self.iface.feedEncodedBytes(data)
                except IOError as e:  # and here it is handeled
                    self.iface.service.print(str(e))
                    break
                    if e.errno == errno.EWOULDBLOCK:
                        pass

class PolyUdp (threading.Thread):
    def __init__(self, iface, localPort ):
        threading.Thread.__init__(self)
        self.iface = iface
        self.localPort = localPort
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.setblocking(0)
        self.socket.bind(("0.0.0.0", self.localPort))
        self.iface.print(" UDP Listening on port: " + str(self.socket.getsockname()[1]))
        self.host = 0

    def __del__(self):
        self.socket.close()
        self.join()

    def close(self):
        self.socket.close()

    def connect(self, hostIp, hostPort):
        self.iface.print("UDP Connecting to " + hostIp + ":"+str(hostPort))
        self.host = (hostIp, hostPort)

    def send(self, data):
        if not self.host == 0:
            try:
                #self.iface.print(str(self.host))
                #self.iface.print(" >>> " + ''.join(' {:02x}'.format(x) for x in data))
                self.socket.sendto(data, self.host)
            except Exception  as e:
                self.iface.print(str(e))

    def run(self):
        while True:
            try:
                data, address = self.socket.recvfrom(1024)
                if data:
                    if self.host == 0:
                        self.host = address
                        self.iface.service.print(" Connection Accepted: " + str(self.host))
                    #self.iface.print(" <<< " + ''.join(' {:02x}'.format(x) for x in data))
                    self.iface.feedEncodedBytes(data)
            except IOError as e:  # and here it is handeled
                if e.errno == errno.EWOULDBLOCK:
                    pass


class PolyField:
    def __init__(self, desc):
        self.desc = desc
        self.id = desc.id
        self.isPresent = False
        self.values = [-1]
        self.len = 0

    def set(self,val):
        self.isPresent = True
        if type(val) != 'str':
            val = str(val)

        if val in self.desc.valDict:
            self.len = 1
            self.values[0] = self.desc.valDict[val]
        else:
            if self.desc.isArray and not self.desc.isString:
                val = val.replace('[','').replace(']','')
                arrVal = val.split(',')
                self.len = len(arrVal)
                self.values = []
                for v in arrVal:
                    self.values.append( int(v, 0))
            else:
                if self.desc.isString:
                    self.len = len(val)
                    self.values[0] = val

                elif self.desc.isMask and '|' in val:
                    self.len = 1
                    self.values[0] = 0
                    arrVals = val.split('|')
                    for v in arrVals:
                        v = v.strip()
                        if v in self.desc.valDict:
                            self.values[0] = self.values[0] | self.desc.valDict[v]
                else:
                    self.len = 1
                    self.values[0] = int(val, 0)


    def get(self):
        if self.isPresent :
            if self.desc.isArray and not self.desc.isString :
                return self.values
            else:
                return self.values[0]
        else:
            return -1

    def parse(self, bytes):
        self.isPresent = True
        strFormat = "%s"+ self.desc.pyFormat
        idx =0
        if self.desc.isArray and not self.desc.isString:
            self.len, idx = readVarSize(bytes)
            self.len = int(self.len / self.desc.objSize)
        else:
            if self.desc.isString:
                self.len, idx = readVarSize(bytes)
            else:
                self.len = 1

        if self.len == 0:
            if self.desc.isString:
                self.values[0] =""
            else:
                self.values[0] = 0

        dataLen = int(self.len * self.desc.objSize)

        strFormat = "<" + str(self.len) + self.desc.pyFormat;

        try:
            if self.desc.isString:
                self.values[0] = struct.unpack(strFormat, bytes[idx:idx+dataLen ])[0].decode("utf-8")
            else:
                self.values = struct.unpack(strFormat, bytes[idx:idx+dataLen])
        except:
            print(strFormat)
            print(idx)
            print (dataLen)
            print(len(bytes))
            print(" Error Parsing: " + self.desc.name+ '-->' + ''.join(' {:02x}'.format(x) for x in bytes[idx:idx+dataLen]))
        return idx + dataLen

    def pack(self, id):
        byteArr = bytes([])

        strFormat = "<" + str(self.len)+ self.desc.pyFormat

        byteArr += packVarSize(id)

        if self.desc.isArray :
            byteArr += packVarSize(self.len * self.desc.objSize)


        if self.desc.isString:
            byteArr+= struct.pack("<" +str(self.len) + "s", self.values[0].encode('utf-8'))
        else:
            byteArr+= struct.pack(strFormat, *self.values)

        return byteArr

    def printJSON(self):
        json =""
        json += "\"" + self.desc.name +"\" : "
        if self.desc.isArray and not self.desc.isString:
            json+= "[" + ''.join(' 0x{:02x},'.format(x) for x in self.values) + "]"
        else:
            if self.desc.isString :
                json+= "\"" + str(self.values[0]) +"\""
            else:
                json+= str(self.values[0])

        return json

class PolyPacket:
    def __init__(self, protocol ):
        self.protocol = protocol
        self.fields = []
        self.seq =0
        self.dataLen = 0
        self.token = random.randint(1, 32767)
        self.checksum = 0
        self.typeId = 0
        self.packet_handler = ''
        self.autoAck = True
        self.ackFlag = False

    def setField(self, fieldName, value):
        for field in self.fields:
            if fieldName.lower() == field.desc.name.lower():
                field.set(value)
                break

    def getField(self, fieldName):
        for field in self.fields:
            if fieldName.lower() == field.desc.name.lower():
                return field.get()
        return -1



    def build(self, typeId):
        self.typeId = typeId
        self.desc = self.protocol.structsAndPackets[typeId]
        for fieldDesc in self.desc.fields:
            self.fields.append( PolyField(fieldDesc))

    def copyTo(self, packet):
        return 0

    def handler(self, iface):

        #dont respond to acks
        if self.ackFlag:
            if self.desc.name in iface.service.handlers:
                iface.service.handlers[self.desc.name](iface.service,self,None )
            elif 'default' in iface.service.handlers:
                iface.service.handlers['default'](iface.service, self,None)
            return 0

        if not iface.service.autoAck:
            return 0

        resp = 0

        if self.desc.hasResponse:
            resp =  iface.service.newPacket(self.desc.response.name)
        elif iface.service.autoAck:
            resp = iface.service.newPacket('Ack')
        
        #acks responding to pings get an icd field
        if self.typeId == 0:
                resp.setField('icd', str(self.protocol.crc))

        if self.desc.name in iface.service.handlers:
            iface.service.handlers[self.desc.name](iface.service,self,resp)
        elif 'default' in iface.service.handlers:
            iface.service.handlers['default'](iface.service, self,resp)


        if not resp == 0:
            resp.ackFlag = True
            resp.token = self.token | 0x8000

        return resp

    def parse(self, rawBytes):
        self.raw = rawBytes
        idx =0
        #pull in header
        self.typeId = rawBytes[0]
        self.seq = rawBytes[1]
        self.dataLen = (rawBytes[3] << 8) | rawBytes[2]
        self.token =   (rawBytes[5] << 8) | rawBytes[4]
        if rawBytes[5] & 0x80:
            self.ackFlag = True
        self.checksum =   (rawBytes[7] << 8) | rawBytes[6]

        idx = 8

        #look up desc
        self.build(self.typeId)

        #parse fields
        while idx < (len(rawBytes)-1):
            fieldId, varLenSize = readVarSize(rawBytes[idx:])
            idx+= varLenSize
            idx+= self.fields[fieldId].parse(rawBytes[idx:])


        return True

    def pack(self):
        byteArr =  bytes([])
        dataArr = bytes([])
        self.checksum = 1738

        #TODO add header
        for i,field in enumerate(self.fields):
            if field.isPresent:
                dataArr += field.pack(i)

        for dat in dataArr:
            self.checksum += dat

        self.dataLen = len(dataArr)

        byteArr = struct.pack('<BBHHH', self.typeId, self.seq, self.dataLen, self.token, self.checksum)
        self.raw = byteArr + dataArr

        return self.raw

    def printJSON(self, meta= False):
        json = ""
        #json += ''.join(' {:02x}'.format(x) for x in self.raw) + "\n"
        json +="{ \"packetType\" : \""+ self.desc.name + "\""

        if meta:
            json+= ", \"typeId\": "  + str(self.typeId)
            json+= ", \"token\": \"" + '{:04x}'.format(self.token) + "\""
            json+= ", \"checksum\": \"" + '{:04x}'.format(self.checksum) + "\""
            json+= ", \"len\": "  + str(self.dataLen) + " "


        for field in self.fields:
            if field.isPresent:
                json+= ", " + field.printJSON()

        json += "}"
        return json

class PolyIface:
    def __init__(self, connStr, service):
        self.connStr = connStr
        self.service = service
        self.bytesIn = deque([])
        self.frameCount =0
        self.packetsIn = deque([])
        self.name = ""
        self.lastToken = 0


        if not connStr == "":
            words = connStr.split(':')  
            localPort = 0    #randomly picks a free port
            remotePort = -1
            remoteHost = '127.0.0.1'

            #udp:local_port:remote_port
            #udp:local_port:remote_ip:remote_port
            #udp:remote_ip:remote_port

            if words[0] == 'udp':
                try:
                    if(words[1].isnumeric()): 
                        localPort = int(words[1])
                        if len(words) == 3:              #udp:local_port:remote_port
                            localPort = int(words[1])
                            remotePort = int(words[2])
                        if len(words) == 4:              #udp:local_port:remote_ip:remote_port
                            remoteHost = words[2]
                            remotePort = int(words[3])
                    else:                                #udp:remote_ip:remote_port
                        remoteHost = words[1]
                        remotePort = int(words[2])

                except:
                    udp_help = ""
                    udp_help+= "Invalid connection string. Options:\n"
                    udp_help+= "\t[udp:local-port] to listen on a port\n"
                    udp_help+= "\t[udp:host:remote-port] to target remote port on a host and use the default local port\n"
                    udp_help+= "\t[udp:local-port:host:remote-port] to target remote port on host and specify local hose\n"
                    udp_help+= "\t[udp:remote-port] to target port on local host\n"
                    udp_help+= "\t[udp:local-port:remote-port] to target port on local host, specifying local port\n"
                    self.print( udp_help)

                self.coms = PolyUdp(self, localPort)
                self.name = "UDP"
                if remotePort > -1:
                    self.coms.connect(remoteHost, remotePort)
                
                self.coms.daemon = True
                self.coms.start()
            
            if words[0] == 'tcp':
                try:
                    if(words[1].isnumeric()):           #tcp:local_port
                        localPort = int(words[1])   
                    else:                                #tcp:remote_ip:remote_port
                        remoteHost = words[1]
                        remotePort = int(words[2])

                except:
                    tcp_help = ""
                    tcp_help+= "Invalid connection string. Options:\n"
                    tcp_help+= "\t[tcp:local-port] to listen on a port\n"
                    tcp_help+= "\t[tcp:host:remote-port] to target remote port on a host and use the default local port\n"
                    tcp_help+= "\t[tcp:local-port:host:remote-port] to target remote port on host and specify local hose\n"
                    tcp_help+= "\t[tcp:remote-port] to target port on local host\n"
                    tcp_help+= "\t[tcp:local-port:remote-port] to target port on local host, specifying local port\n"
                    self.print( tcp_help)

                self.coms = PolyTcp(self, localPort)
                self.name = "TCP"
                if remotePort == -1:    #server mode
                    self.coms.listen()
                else:                   #client mode 
                    self.coms.connect(remoteHost, remotePort)
                
                self.coms.daemon = True
                self.coms.start()

            elif words[0] == 'serial':
                try:
                    if len(words) == 2:
                        self.coms = PolySerial(self,words[1], 9600)
                    if len(words) == 3:
                        self.coms = PolySerial(self,words[1], words[2])

                    self.coms.daemon = True
                    self.coms.start()
                except:
                    self.print( "Invalid connection string\n serial:/dev/ttyS[COM Number]:baud")
                    
    def close(self):
        if hasattr(self, 'coms'):
            self.coms.close()

    def print(self, text):
        if not self.service.print == '':
            self.service.print( text)

    def feedEncodedBytes(self, encodedBytes):

        silent = False

        for i in encodedBytes:
            self.bytesIn.append(i)
            if i == 0:
                self.frameCount +=1

        while self.frameCount > 0:
            encodedPacket = bytes([])
            newPacket = PolyPacket(self.service.protocol)
            while(1):
                x = self.bytesIn.popleft()
                if x == 0:
                    self.frameCount -=1
                    break
                else:
                    encodedPacket+= bytes([x])

            decoded = cobs.decode(encodedPacket)

            try:
                newPacket.parse(decoded)
            except Exception  as e:
                self.print(str(e))
                
            if self.service.silenceDict[newPacket.desc.name]:
                silent = True

            if not silent:
                if self.service.showBytes:
                    self.print(" PARSE HDR: " + ''.join(' {:02x}'.format(x) for x in decoded[:8]))
                    self.print(" PARSE DATA: " + ''.join(' {:02x}'.format(x) for x in decoded[8:]))
                if (newPacket.token & 0x7FFF) != (self.lastToken & 0x7FFF):
                    self.print("")

                self.print( " <-- " + newPacket.printJSON(self.service.showMeta))

            resp = newPacket.handler(self)
            self.lastToken = newPacket.token
            if resp:
                self.sendPacket(resp, silent)
            #self.packetsIn.append(newPacket)

    def sendPacket(self, packet, silent = False):

        if packet.desc.name == "Ping":
            packet.setField('icd', str(self.service.protocol.crc))

        if (packet.token & 0x7FFF) != (self.lastToken & 0x7FFF):
            self.print("")

        if not silent:
            self.print( " --> " + packet.printJSON(self.service.showMeta))

        raw = packet.pack()

        if self.service.showBytes and not silent:
            self.print(" PACK HDR: " + ''.join(' {:02x}'.format(x) for x in raw[:8]))
            self.print(" PACK DATA: " + ''.join(' {:02x}'.format(x) for x in raw[8:]))

        encoded = cobs.encode(bytearray(raw))
        encoded += bytes([0])


        if hasattr(self, 'coms'):
            self.coms.send(encoded)

        self.lastToken = packet.token

        return encoded


    def getPacket(self):
        if len(packetsIn) > 0:
            return packetsIn.popleft()



class PolyService:
    def __init__(self, protocol):
        self.protocol = protocol
        self.interfaces = []
        self.print = ''
        self.showMeta = False
        self.autoAck = True
        self.handlers = {}
        self.silenceDict = {}
        self.showBytes = False
        self.dataStore = {}

        for packet in protocol.packets:
            self.silenceDict[packet.name] = False

        self.addIface("") #add dummy interface that just prints

    def close(self):
        for iface in self.interfaces:
            iface.close()

    def addIface(self, connStr):
        self.interfaces.append(PolyIface(connStr, self))

    def setIface(self,connStr):
        self.interfaces[0].close()
        self.interfaces[0] = PolyIface(connStr, self)

    def toggleAck(self):
        if self.autoAck:
            self.autoAck = False
            self.print( "AutoAck turned OFF")
        else:
            self.autoAck = True
            self.print( "AutoAck turned ON")

    def toggleSilence(self, packetType):

        if not packetType in self.silenceDict:
            self.print( "Can not find: " + packetType)

        if self.silenceDict[packetType]:
            self.silenceDict[packetType] = False
            self.print( "Un-Silencing: " + packetType)
        else:
            self.silenceDict[packetType] = True
            self.print( "Silencing: " + packetType)

    def newPacket(self, type):
        packet = PolyPacket(self.protocol)

        if type in self.protocol.packetIdx:
            packet.build(self.protocol.packetIdx[type])
        else:
            self.print(" Packet Type \"" + type + "\", not found!")

        return packet

    def newStruct(self, type):
        return self.newPacket(type)
    #
    # def process(self):
    #     for iface in self.interfaces:
    #         if iface.frameCount > 0:



import sys
import xml.etree.ElementTree as ET
import re
import io
import os
import copy
import datetime
import zlib
import argparse
#import pdfkit
from shutil import copyfile
from mako.template import Template
import pkgutil
import polypacket
import subprocess
import yaml


sizeDict = {
    "uint8" : 1,
    "int8" : 1,
    "char" : 1,
    "string" : 1,
    "uint16" : 2,
    "int16" : 2,
    "uint32" : 4,
    "int32" : 4,
    "int64" : 8,
    "uint64" : 8,
    "int" : 4,
    "float": 4,
    "double": 8,
}

cNameDict = {
    "uint8" : "uint8_t",
     "int8" : "int8_t",
     "char" : "char",
     "string" : "char",
     "uint16" : "uint16_t",
     "int16" : "int16_t",
     "uint32" : "uint32_t",
     "int32" : "int32_t",
     "int64" : "int64_t",
     "uint64" : "uint64_t",
     "int" : "int",
     "float" : "float",
     "double" : "double",
     "enum" : "uint8_t",
     "flag" : "uint8_t"
 }

formatDict = {
    "hex" : "FORMAT_HEX",
    "dec" : "FORMAT_DEC",
    "default" : "FORMAT_DEFAULT",
    "ascii" : "FORMAT_ASCII",
    "none" : "FORMAT_NONE"
}

pyFormatDict = {
        "uint8" : "B",
        "int8" : "b",
        "char" : "c",
        "string" : "s",
        "uint16" : "H",
        "int16" : "h",
        "uint32" : "L",
        "int32" : "l",
        "int64" : "q",
        "uint64" : "Q",
        "int" : "l",
        "float": "f",
        "double": "d",
}


def crc(fileName):
    prev = 0
    for eachLine in open(fileName,"rb"):
        prev = zlib.crc32(eachLine, prev)


    return prev,"%X"%(prev & 0xFFFFFFFF)

class simulator:
    def __init__(self,name, simItem):
        self.init =""
        self.handlers = {}
        self.name = name
        
        if 'init' in simItem:
            self.init = simItem['init']

        if 'handlers' in simItem:
            for handler in simItem['handlers']:
                name = list(handler.keys())[0]
                code  = list(handler.values())[0]
                self.handlers[name] = code

class fieldVal:
    def __init__(self, name):
        self.name = name.upper()
        self.desc = ""
        self.val = None
        

class fieldDesc:
    def __init__(self, name, strType):
        self.vals = []
        self.valDict = {}
        self.arrayLen = 1
        self.isEnum = False
        self.isMask = False
        self.valsFormat = "0x%0.2X"
        self.valIndex = 0

        self.format = 'FORMAT_DEFAULT'

        if strType in ['flag','flags','mask','bits']:
            self.format = 'FORMAT_HEX'
            self.isMask = True
            strType = 'uint8_t'

        if strType in ['enum','enums']:
            self.format = 'FORMAT_HEX'
            self.isEnum = True
            strType = 'uint8_t'

        m = re.search('\*([0-9]*)', strType)
        if(m):
            if(m.group(1) != ''):
                self.arrayLen = int(m.group(1))
            strType = strType[0:m.start()]

        strType = strType.lower().replace('_t','')

        self.setType(strType, self.arrayLen)

        self.id = 0
        self.name = name
        self.globalName = "PP_FIELD_"+self.name.upper()
        self.isVarLen = False
        self.isRequired = False
        self.desc = ""
        self.memberName = "m"+ self.name.capitalize()

    def camel(self):
        return self.name[:1].capitalize() + self.name[1:]

    def setType(self, type, len):

        if not (type in cNameDict):
            print( "INVALID DATA TYPE!:  " + type)

        self.arrayLen = len
        self.type = type
        self.size = sizeDict[self.type] * self.arrayLen
        self.objSize = sizeDict[self.type]
        self.pyFormat = pyFormatDict[self.type]
        self.cType = cNameDict[self.type]
        self.cppType = self.cType

        self.isString = False
        self.isArray = False

        if(self.arrayLen > 1):
            self.isArray = True

        if(self.type == 'string'):
            self.cppType = "string"
            self.isString = True
            self.isArray = True
            if(self.arrayLen == 1):
                self.arrayLen = 32 #if no arraylen is specified default 32
        else:
            if(self.isArray):
                self.cppType = self.cppType +"*"

    def addVal(self, val):
 
        self.valDict[val.name] = len(self.vals) -1

        if self.isMask:
            val.val = 1 << self.valIndex
            self.valIndex+=1
            strType = 'uint8'
            if len(self.vals) > 8:
                self.valsFormat = "0x%0.4X"
                strType = 'uint16'
            if len(self.vals) > 16:
                self.valsFormat = "0x%0.8X"
                strType = 'uint32'
            if len(self.vals) > 32:
                print( "Error maximum flags per field is 32")
            self.setType(strType,1)
        
        elif self.isEnum:
            val.val =  self.valIndex
            self.valIndex+=1

        self.valDict[val.name] = val.val
        self.vals.append(val)
    

    def setPrefix(self, prefix):
        self.globalName = prefix.upper()+"_FIELD_"+self.name.upper()

    def getFieldDeclaration(self):
        output = io.StringIO()
        output.write("{0} field_{1}".format(self.cType, self.name))
        if(self.arrayLen > 1):
            output.write("["+str(self.arrayLen)+"]")

        return output.getvalue()

    def getParamType(self):
        if self.isArray:
            return self.cType +"*"
        else:
            return self.cType;

    def getDeclaration(self):
        if self.isArray:
            return self.cType +" "+self.name+"["+ str(self.arrayLen)+"]"
        else:
            return self.cType + " " + self.name;

    def getFormat(self):
        if self.isString:
            return "%s"
        else:
            return "%i"



class packetDesc:
    def __init__(self,name, protocol):
        self.name = name
        self.globalName =  "PP_PACKET_"+self.name.upper()
        self.className = name.capitalize() +"Packet"
        self.desc =""
        self.fields = []
        self.sruct = False
        self.fieldCount=0
        self.respondsTo = {}
        self.requests = {}
        self.standard = False
        self.structName = name.lower() + '_packet_t'
        self.hasResponse = False
        self.protocol = protocol
        self.requiredFields = []
        self.requiredFieldCount = 0

    def camel(self):
        return self.name[:1].capitalize() + self.name[1:]

    def setPrefix(self, prefix):
        self.globalName = prefix.upper()+"_PACKET_"+self.name.upper()

    def addField(self, field):
        field.id = self.fieldCount
        self.fields.append(field)
        self.fieldCount+=1

    def addYAMLField(self, pfieldItem):

        if type(pfieldItem) is dict:
            pfname = list(pfieldItem.keys())[0]
            pfield = list(pfieldItem.values())[0]
        else:
            pfname = pfieldItem
            pfield = {}

        strReq =""
        if not (pfname in self.protocol.fieldIdx):
            print( 'ERROR Field not declared: ' + pfname)

        #get id of field and make a copy
        idx = self.protocol.fieldIdx[pfname]
        fieldCopy = copy.copy(self.protocol.fields[idx])

        if('req' in pfield):
            fieldCopy.isRequired = pfield['req']

        if('desc' in pfield):
            fieldCopy.desc = pfield['desc']

        fieldCopy.id = self.fieldCount
        self.fields.append(fieldCopy)
        self.fieldCount+=1

    def postProcess(self):
        if len(self.requests) > 0:
            self.hasResponse = True;
            self.response = self.protocol.getPacket(next(iter(self.requests.keys())))

        for field in self.fields:
            if field.isRequired:
                self.requiredFields.append(field)
                self.requiredFieldCount += 1

    def tableSize(self):
        sum =0;
        for field in self.fields:
            if field.size > 4:
                sum+=4
            else:
                sum += field.size

        return sum


    def getDocMd(self):
        output = io.StringIO()
        idHex = "%0.2X" % self.packetId
        output.write('### '  + self.name + '\n')
        output.write(self.desc + '\n\n')
        output.write('* Packet ID: *['+idHex+']*\n')
        requestCount = len(self.requests)
        respondsToCount = len(self.respondsTo)

        #write response packets
        if(requestCount > 0):
            output.write('* *Requests: ')
            first = True
            for req in self.requests:
                if(first):
                    first = False
                else:
                    output.write(', ')
                output.write(req)
            output.write('*\n\n')

        #write request packets
        if(self.name == 'Ack'):
            output.write('* *Responds To: Any Packet without a defined response*\n\n')
        else:
            if(respondsToCount > 0):
                output.write('* *Responds To: ')
                first = True
                for resp in self.respondsTo:
                    if(first):
                        first = False
                    else:
                        output.write(', ')
                    output.write(resp)
                output.write('*\n')

        output.write('\n')

        rowBytes = io.StringIO()
        rowBorder = io.StringIO()
        rowFields = io.StringIO()
        rowTypes = io.StringIO()


        if(len(self.fields) > 0):
            rowBytes.write('|***Byte***|')
            rowBorder.write('|---|')
            rowFields.write('|***Field***')
            rowTypes.write('|***Type***')

            count =0

            for pfield in self.fields:

                #write bytes
                if(pfield.size > 4):
                    rowBytes.write(str(count)+'| . . . . . . . |'+str(count+pfield.size -1))
                    count+=pfield.size
                else:
                    for x in range(pfield.size):
                        rowBytes.write(str(count) + '|')
                        count+=1

                #write border
                span = pfield.size
                if(span > 4):
                    span = 4
                for x in range(span):
                    rowBorder.write('---|')

                #write fields
                span = pfield.size
                if(span > 4):
                    span = 4
                rowFields.write('<td colspan=\''+str(span)+'\'>')
                if(pfield.isRequired):
                    rowFields.write('***'+pfield.name+'***')
                else:
                    rowFields.write(pfield.name)

                #write types
                span = pfield.size
                if(span > 4):
                    span = 4
                rowTypes.write('<td colspan=\''+str(span)+'\'>')
                rowTypes.write(pfield.cType)
                if(pfield.isArray):
                    if(pfield.isVarLen):
                        rowTypes.write('[0-'+ str(pfield.size)+' ]')
                    else:
                        rowTypes.write('['+str(pfield.size)+']')

            #combine rows for table
            output.write(rowBytes.getvalue() + "\n");
            output.write(rowBorder.getvalue() + "\n");
            output.write(rowFields.getvalue() + "\n");
            output.write(rowTypes.getvalue() + "\n");

            output.write('\n\n')
            output.write('Fields:\n')
            #write field description table
            for pfield in self.fields:
                output.write('>***'+ pfield.name+'*** : ' + pfield.desc +'<br/>\n')
                if pfield.isMask:
                    for idx,val in enumerate(pfield.vals):
                        strVal = pfield.valsFormat % (1 << idx)
                        output.write('>> **{0}** : {1} - {2}<br/>\n'.format(strVal, val.name, val.desc))
                    output.write('>\n')

                if pfield.isEnum:
                    for idx,val in enumerate(pfield.vals):
                        strVal = pfield.valsFormat % (idx)
                        output.write('>> **{0}** : {1} - {2}<br/>\n'.format(strVal, val.name, val.desc))
                    output.write('>\n')
        else:
            output.write('>This Packet type does not contain any data fields\n\n')

        output.write('\n------\n')

        return output.getvalue();


class protocolDesc:
    def __init__(self, name):
        self.name = name
        self.fileName = name+"Service"
        self.cppFileName = name+"Service"
        self.desc = ""
        self.hash = ""
        self.fields = []
        self.fieldIdx = {}
        self.fieldId =0
        self.fieldGroups = {}
        self.packets = []
        self.packetIdx ={}
        self.packetId =0
        self.structs =[]
        self.structsAndPackets=[]
        self.structIdx ={}
        self.structId =0
        self.prefix = "pp"
        self.snippets = False
        self.genUtility = False
        self.xmlName =""
        self.utilName =""
        self.sims = {}
        self.defaultResponse = ""

    def service(self):
        return self.prefix.upper() +'_SERVICE'

    def descFromId(self, typeId):
        return self.packets[typeId-len(self.structs)]

    def fieldDescFromId(self, typeId):
        return self.fields[typeId]

    def camelPrefix(self):
        return self.prefix[:1].capitalize() + self.prefix[1:]

    def addField(self,field):
        field.id = self.fieldId
        field.protocol = self
        self.fields.append(field)
        self.fieldIdx[field.name] = self.fieldId
        self.fieldId+=1

    def addGroup(self, name, fields):
        self.fieldGroups[name] = fields

    def addPacket(self,packet):
        packet.packetId = self.packetId
        packet.protocol = self
        packet.setPrefix(self.prefix)
        self.packets.append(packet)
        self.structsAndPackets.append(packet)
        self.packetIdx[packet.name] = self.packetId
        self.packetId+=1

    def addStruct(self,struct):
        struct.packetId = self.packetId
        struct.protocol = self
        struct.struct = True
        struct.globalName = self.prefix.upper()+"_STRUCT_"+struct.name.upper()
        self.structs.append(struct)
        self.structsAndPackets.append(struct)
        self.structIdx[struct.name] = self.packetId
        self.packetId+=1

    def getPacket(self, name):
        if name in self.packetIdx:
            return self.structsAndPackets[self.packetIdx[name]]




def addStandardPackets(protocol):
    ping = packetDesc("Ping", protocol)
    ack = packetDesc("Ack", protocol)
    icd = fieldDesc("icd", "uint32")
    icd.isRequired = True
    icd.format = 'FORMAT_HEX'
    icd.setPrefix(protocol.prefix)
    icd.desc = "CRC Hash of protocol description. This is used to verify endpoints are using the same protocol"
    ping.desc = "This message requests an Ack from a remote device to test connectivity"
    ping.response = ack
    ping.hasResponse = True
    ping.requests['Ack'] =0
    ack.desc ="Acknowledges any packet that does not have an explicit response"
    ping.standard = True
    ack.standard = True
    protocol.addField(icd)
    ping.addField(icd)
    ack.addField(icd)
    protocol.addPacket(ping)
    protocol.addPacket(ack)

def parseXML(xmlfile):

    # create element tree object
    tree = ET.parse(xmlfile)

    # get root element
    root = tree.getroot()

    # create empty list for Fields
    protocol = protocolDesc(root.attrib['name'])
    protocol.xmlName = os.path.basename(xmlfile)


    if('desc' in root.attrib):
        protocol.desc = root.attrib['desc']

    if('prefix' in root.attrib):
        protocol.prefix = root.attrib['prefix']

    addStandardPackets(protocol)

    #parse out fields
    for field in root.findall('./Fields/Field'):

        name = field.attrib['name']
        strType = field.attrib['type'];

        newField = fieldDesc(name, strType)
        newField.setPrefix(protocol.prefix)

        if('format' in field.attrib):
            format = field.attrib['format'].lower()
            if not format in formatDict:
                print( "INVALID FORMAT :" + format)

            newField.format = formatDict[format]

        if('desc' in field.attrib):
            newField.desc = field.attrib['desc']

        if(name in protocol.fields):
            print( 'ERROR Duplicate Field Name!: ' + name)

        #get vals if any
        for val in field.findall('./Val'):
            name = val.attrib['name']
            newVal = fieldVal(name)

            if('desc' in val.attrib):
                newVal.desc = val.attrib['desc']

            newField.addVal(newVal)



        protocol.addField(newField)


    #get all packet types
    for packet in root.findall('./Packets/Packet'):
        name = packet.attrib['name']
        desc =""
        newPacket = packetDesc(name, protocol)
        newPacket.setPrefix(protocol.prefix)

        if(name in protocol.packetIdx):
            print( 'ERROR Duplicate Packet Name!: ' + name)

        if('desc' in packet.attrib):
            desc = packet.attrib['desc']

        if('response' in packet.attrib):
            newPacket.requests[packet.attrib['response']] = 0

        #get all fields declared for packet
        for pfield in packet:

            pfname = pfield.attrib['name']
            strReq =""
            if not (pfname in protocol.fieldIdx):
                print( 'ERROR Field not declared: ' + pfname)

            #get id of field and make a copy
            idx = protocol.fieldIdx[pfname]
            fieldCopy = copy.copy(protocol.fields[idx])

            if('req' in pfield.attrib):
                strReq = pfield.attrib['req']
                if(strReq.lower() == "true" ):
                    fieldCopy.isRequired = True

            if('desc' in pfield.attrib):
                fieldCopy.desc = pfield.attrib['desc']

            newPacket.addField(fieldCopy)

        newPacket.desc = desc

        protocol.addPacket(newPacket)

    #get all packet types
    for struct in root.findall('./Structs/Struct'):
        name = struct.attrib['name']
        desc =""
        newStruct = packetDesc(name, protocol)


        if(name in protocol.structIdx):
            print( 'ERROR Duplicate Struct Name!: ' + name)

        if('desc' in packet.attrib):
            desc = packet.attrib['desc']

        #get all fields declared for packet
        for pfield in struct:

            pfname = pfield.attrib['name']
            strReq =""
            if not (pfname in protocol.fieldIdx):
                print( 'ERROR Field not declared: ' + pfname)

            #get id of field and make a copy
            idx = protocol.fieldIdx[pfname]
            fieldCopy = copy.copy(protocol.fields[idx])


            if('desc' in pfield.attrib):
                fieldCopy.desc = pfield.attrib['desc']

            newStruct.addField(fieldCopy)

        newStruct.desc = desc

        protocol.addStruct(newStruct)


    for packet in protocol.packets:
        for request in packet.requests:
            idx = protocol.packetIdx[request]
            protocol.packets[idx].respondsTo[packet.name] = 0

    for packet in protocol.packets:
        packet.postProcess()


    # return news items list
    return protocol


def parseYAMLField(protocol, fieldItem):
    name = list(fieldItem.keys())[0]
    field = list(fieldItem.values())[0]

    strType = field['type'].replace("(","[").replace(")","]");

    newField = fieldDesc(name, strType)
    newField.setPrefix(protocol.prefix)

    if('format' in field):
        format = field['format'].lower()
        if not format in formatDict:
            print( "INVALID FORMAT :" + format)

        newField.format = formatDict[format]

    if 'req' in field:
        newField.isRequired = field['req']

    if 'required' in field:
        newField.isRequired = field['required']

    if('desc' in field):
        newField.desc = field['desc']

    if(name in protocol.fields):
        print( 'ERROR Duplicate Field Name!: ' + name)

    #get vals if any
    if "vals" in field:
        for valItem in field['vals']:
            if type(valItem) is dict:
                name = list(valItem.keys())[0]
                val = list(valItem.values())[0]
            else:
                name = valItem
                val = {}

            newVal = fieldVal(name)

            if('val' in val):
                newVal.val = val['val']

            if('desc' in val):
                newVal.desc = val['desc']

            newField.addVal(newVal)

    protocol.addField(newField)
    return newField

def parseYAML(yamlFile):
    data = open(yamlFile)
    objProtocol = yaml.load(data , Loader=yaml.FullLoader)

    protocol = protocolDesc(objProtocol['name'])

    if "prefix" in objProtocol:
        protocol.prefix = objProtocol['prefix']

    if "desc" in objProtocol:
        protocol.desc = objProtocol['desc']

    if "defaultResponse" in objProtocol:
        protocol.defaultResponse = objProtocol['defaultResponse']

    addStandardPackets(protocol)

    protocol.xmlName = os.path.basename(yamlFile)

    for fieldItem in objProtocol['fields']:

        nodeType = list(fieldItem.values())[0]

        #all fields must have a 'type', so if it doesnt, then it is a field group
        if not 'type' in list(fieldItem.values())[0]:
            groupName = list(fieldItem.keys())[0]
            fieldGroupItems = list(fieldItem.values())[0]
            groupFields = []
            for fieldGroupItem in fieldGroupItems:
                newField = parseYAMLField(protocol, fieldGroupItem)
                groupFields.append(newField.name)
            protocol.addGroup(groupName, groupFields)
        else:
            parseYAMLField(protocol, fieldItem)

    if 'structs' in  objProtocol:
        for structItem in objProtocol['structs']:
            name = list(structItem.keys())[0]
            struct = list(structItem.values())[0]
            desc =""
            newStruct = packetDesc(name,protocol)


            if(name in protocol.structIdx):
                print( 'ERROR Duplicate Struct Name!: ' + name)

            if('desc' in struct):
                desc = struct['desc']

            #get all fields declared for packet
            if "fields" in struct:
                for pfieldItem in struct['fields']:

                    if type(pfieldItem) is dict:
                        pfname = list(pfieldItem.keys())[0]
                        pfield = list(pfieldItem.values())[0]
                    else:
                        pfname = pfieldItem
                        pfield = {}


                    if pfname in protocol.fieldGroups:
                        for pfFieldGroupItem in protocol.fieldGroups[pfname]:
                            newStruct.addYAMLField(pfFieldGroupItem)
                    else:
                        newStruct.addYAMLField(pfieldItem)

            newStruct.desc = desc

            protocol.addStruct(newStruct)

    if 'packets' in  objProtocol:
        for packetItem in objProtocol['packets']:
            name = list(packetItem.keys())[0]
            packet = list(packetItem.values())[0]
            desc =""
            newPacket = packetDesc(name, protocol)
            newPacket.setPrefix(protocol.prefix)

            if(name in protocol.packetIdx):
                print( 'ERROR Duplicate Packet Name!: ' + name)

            if('desc' in packet):
                desc = packet['desc']

            if('response' in packet):
                if (packet['response'] != "none"):
                    newPacket.requests[packet['response']] = 0
            else:
                if not protocol.defaultResponse == "" and not  protocol.defaultResponse == newPacket.name :
                    newPacket.requests[protocol.defaultResponse] = 0

            #get all fields declared for packet
            if "fields" in packet:
                for pfieldItem in packet['fields']:

                    if type(pfieldItem) is dict:
                        pfname = list(pfieldItem.keys())[0]
                        pfield = list(pfieldItem.values())[0]
                    else:
                        pfname = pfieldItem
                        pfield = {}


                    if pfname in protocol.fieldGroups:
                        for pfFieldGroupItem in protocol.fieldGroups[pfname]:
                            newPacket.addYAMLField(pfFieldGroupItem)
                    else:
                        newPacket.addYAMLField(pfieldItem)

            newPacket.desc = desc

            protocol.addPacket(newPacket)


    if 'sims' in  objProtocol: #experimental
        for simItem in objProtocol['sims']:
            name = list(simItem.keys())[0]
            sim = list(simItem.values())[0]
            protocol.sims[name] = simulator(name,sim)
    
    for packet in protocol.packets:
        for request in packet.requests.keys():
            protocol.getPacket(request).respondsTo[packet.name] = 0

    for packet in protocol.packets:
        packet.postProcess()


    # return news items list
    return protocol

def buildProtocol(file):
    extension = os.path.splitext(file)[1]

    if(extension == ".xml"):
        print(" XML files are depreciated. Please convert to YAML for future use")
        return parseXML(file)

    elif(extension == ".yml"):
        return parseYAML(file)

    else:
        print(" Files Type: " + extension+" Not supported. Please use YAML")

    return 0



import setuptools




with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
     name='polypacket',
     version='1.0.61',
     author="Jason Berger",
     author_email="JBerger@up-rev.com",
     description="A tool for building protocol services",
     long_description=long_description,
     scripts=['polypacket/poly-packet','polypacket/poly-make'],
     long_description_content_type="text/markdown",
     url="http://www.up-rev.com/",
     packages=setuptools.find_packages(),
     package_data={
     'polypacket':['templates/*','examples/*']
     },
     install_requires=[
        'markdown',
        'mako',
        'prompt_toolkit',
        'pyyaml',
        'cobs',
        'pyserial',
        'update_notipy'
     ],
     classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
     ],
 )
